
/**
 *	Reads run-length-encoded images from an InputStream and reports the
 *	image data to an InputStreamImageSource object.
 */

import sun.awt.image.*;
import java.awt.image.*;

import java.util.Vector;
import java.util.Hashtable;
import java.io.InputStream;
import java.io.IOException;

/**
 * Gif Image converter
 * 
 * @version 1.0 07/03/97
 * @author Jeff Nielsen 
 */
public class RLImageDecoder extends ImageDecoder {
    // All I need to implement is catchupConsumer() and produceImage()

    private static final boolean verbose = false;


    private static final int DISPOSAL_NONE      = 0x00;
    private static final int DISPOSAL_SAVE      = 0x01;
    private static final int DISPOSAL_BGCOLOR   = 0x02;
    private static final int DISPOSAL_PREVIOUS  = 0x03;

    private static final int TAG4 = 0x84; 
    private static final int TAG8 = 0x88; 
    private static final int TAG12 = 0x8c; 
    private static final int TAG16 = 0x90; 


    
    private static IndexColorModel trans_model;

    PixelStore8 store;
    boolean cancatchup = true;

//    int num_global_colors;
//    byte[] global_colormap;
//    int trans_pixel = -1;
//    IndexColorModel global_model;
    static ColorModel global_model;

    Hashtable props = new Hashtable();

    byte[] saved_image;
//    ColorModel saved_model;
    int disposal_method = DISPOSAL_NONE;

    private int tag;

    int global_width;
    int global_height;
    int global_bgpixel;

//    boolean firstframe = true;
//    int nloops = -1;
//    int delay = -1;

    public RLImageDecoder(InputStreamImageSource src, InputStream is) {
    	super(src, is);
    }

    public static void setColorModel(ColorModel cm) { global_model = cm; }

    public synchronized boolean catchupConsumer(InputStreamImageSource src,
						ImageConsumer ic) {
	    return (cancatchup && (store == null || store.replay(src, ic)));
    }

    public synchronized void makeStore(int width, int height) {
    	if (cancatchup) {
	        store = new PixelStore8(width, height);
	    }
    }

    /**
     * An error has occurred. Throw an exception.
     */
    private static void error(String s1) throws ImageFormatException {
    	throw new ImageFormatException(s1);
    }

    /**
     * Read a number of bytes into a buffer.
     * @return number of bytes that were not read due to EOF or error
     */
    private int readBytes(byte buf[], int off, int len) {
	    while (len > 0) {
	        try {
		        int n = input.read(buf, off, len);
		        if (n < 0) {
		            break;
		    }
		        off += n;
		        len -= n;
	        } catch (IOException e) {
    		    break;
	        }
	    }
	    return len;
    }

    private static final int ExtractByte(byte buf[], int off) {
	    return (buf[off] & 0xFF);
    }

    private static final int ExtractWord(byte buf[], int off) {
    	return (buf[off] & 0xFF) | ((buf[off + 1] & 0xFF) << 8);
    }

    /**
     * produce an image from the stream.  
     * What does it mean to "produce" it?
     */
    public void produceImage() throws IOException, ImageFormatException {
    	try {
    	    readHeader();       


/*            
            int totalframes = 0;
            int frameno = 0;

	        while (true) {

		        int code;

		        switch (code = input.read()) {
		          case EXBLOCK:
		            switch (code = input.read()) {
		              case EX_GRAPHICS_CONTROL: {
			            byte buf[] = new byte[6];
			            if (readBytes(buf, 0, 6) != 0) {
			                return;//error("corrupt GIF file");
			            }
			            if ((buf[0] != 4) || (buf[5] != 0)) {
			                return;//error("corrupt GIF file (GCE size)");
			            }
			            // Get the index of the transparent color
			            delay = ExtractWord(buf, 2) * 10;
			            if (delay > 0) {
			                cancatchup = false;
			                store = null;
			                ImageFetcher.startingAnimation();
			            }
			            disposal_method = (buf[1] >> 2) & 7;
			            if ((buf[1] & TRANSPARENCYMASK) != 0) {
			                trans_pixel = ExtractByte(buf, 4);
			            } else {
			                trans_pixel = -1;
			            }
			            break;
		              }

		              case EX_COMMENT:
		              case EX_APPLICATION:
		              default:
			            boolean loop_tag = false;
			            String comment = "";
			            while (true) {
			                int n = input.read();
			                if (n <= 0) {
				                break;
			                }
			                byte buf[] = new byte[n];
			                if (readBytes(buf, 0, n) != 0) {
				                return;//error("corrupt GIF file");
			                }
			                if (code == EX_COMMENT) {
				                comment += new String(buf, 0);
			                } else if (code == EX_APPLICATION) {
				                if (loop_tag) {
				                    if (n == 3 && buf[0] == 1) {
					                    nloops = ExtractWord(buf, 1);
				                    } else {
					                    loop_tag = false;
				                    }
				                }
				                if ("NETSCAPE2.0".equals(new String(buf, 0))) {
				                    loop_tag = true;
				                }
			                }
			            }   // end while (true)
			            if (code == EX_COMMENT) {
			                props.put("comment", comment);
			            }
			            if (loop_tag) {
			                input.mark(1000000);
			                cancatchup = false;
			                store = null;
			                ImageFetcher.startingAnimation();
			            }
			            break;

		              case -1:
			            return; //error("corrupt GIF file");
		            }   // end switch (code = input.read());
		            break;  // end case EXBLOCK

		          case IMAGESEP:
		            try {

*/
                        // Following line is where image is read!!
            readImage(true);
/*
			            if (!readImage(totalframes == 0)) { 
			                cancatchup = false;
			                return;
    			        }
		            } catch (Exception e) {
			            if (verbose) {
			                e.printStackTrace();
    			        }
	    		        return;
		            }
		            frameno++;
		            totalframes++;
		            firstframe = false;
		            break;

		          default:
		          case -1:
		            if (verbose) {
			            if (code == -1) {
			                System.err.println("Premature EOF in GIF file," +
					                   " frame " + frameno);
			            } else {
			                System.err.println("corrupt GIF file (parse) ["
					                   + code + "].");
			            }
		            }
		            if (frameno == 0) {
    			        return;
		            }
		            // NOBREAK

		          case TERMINATOR:
		            if (nloops == 0 || nloops-- >= 0) {
			            try {
			                input.reset();
			                saved_image = null;
			                saved_model = null;
			                firstframe = true;
			                frameno = 0;
			                break;
			            } catch (IOException e) {
			            }
		            }
		            if (verbose && frameno != 1) {
			            System.out.println("processing GIF terminator,"
					               + " frames: " + frameno
					               + " total: " + totalframes);
		            }
*/
		            if (store != null) {
			            store.imageComplete();
			            if (store.getBitState() != PixelStore.BITS_LOST) {
                            // Make sure to set the pixel store
			                source.setPixelStore(store);
			            }
		            }
		            source.imageComplete(ImageConsumer.STATICIMAGEDONE);
/*
		            return;
		        }   // end switch
            }   // end while (true)
*/
	    } finally {
	        try {
    		    input.close();
	        } catch (IOException e) {
	        }
	    }
    }

    /**
     * Read Image header
     */
    private void readHeader() throws IOException, ImageFormatException {
	    // Create a buffer
	    byte buf[] = new byte[12];

	    // Read the header
	    if (readBytes(buf, 0, 13) != 0) {
	        throw new IOException();
	    }

	    // Check header
	    if ((buf[0] != 'R') || (buf[1] != 'U') || (buf[2] != 'N')) {
	        error("not a RUN file.");
	    }

        tag = ExtractByte(buf, 3);

	    // Global width&height
	    global_width = ExtractWord(buf, 4);
	    global_height = ExtractWord(buf, 6);

/*
	    // colormap info
	    int ch = ExtractByte(buf, 10);
	    if ((ch & COLORMAPMASK) == 0) {
	        error("no global colormap in GIF file.");
	    }
	    num_global_colors = 1 << ((ch & 0x7) + 1);

	    global_bgpixel = ExtractByte(buf, 11);

	    if (buf[12] != 0) {
	        props.put("aspectratio", ""+((ExtractByte(buf, 12) + 15) / 64.0));
	    }

	    // Read colors
	    global_colormap = new byte[num_global_colors * 3];
	    if (readBytes(global_colormap, 0, num_global_colors * 3) != 0) {
	        throw new IOException();
	    }
*/
        // Now skip to first data byte
        int toSkip = ExtractWord(buf, 8) - 10;
        while (toSkip > 0) toSkip -= input.skip(toSkip);

    }

    /**
     * The ImageConsumer hints flag 
     */
    private static final int hintsflags =
	    ImageConsumer.TOPDOWNLEFTRIGHT | ImageConsumer.COMPLETESCANLINES |
	    ImageConsumer.SINGLEPASS | ImageConsumer.SINGLEFRAME;


//    private short prefix[]  = new short[4096];
//    private byte  suffix[]  = new byte[4096];
//    private byte  outCode[] = new byte[4097];

/*
    // Sun's code to read GIFs & JPEGs is actually native code!!  That's
    // how it runs so fast.
    private native boolean parseImage(int x, int y, int width, int height,
				      boolean interlace, int initCodeSize,
				      byte block[], byte rasline[],
				      IndexColorModel model);
*/


    /**
     * This method called by the native parseImage() method
     */
/*
    private int sendPixels(int x, int y, int width, int height,
			   byte rasline[], ColorModel model) {
	    int rasbeg, rasend, x2;
	    if (y < 0) {
	        height += y;
	        y = 0;
	    }
	    if (y + height > global_height) {
	        height = global_height - y;
	    }
	    if (height <= 0) {
	        return 1;
	    }
	    // rasline[0]     == pixel at coordinate (x,y)
	    // rasline[width] == pixel at coordinate (x+width, y)
	    if (x < 0) {
	        rasbeg = -x;
	        width += x;		// same as (width -= rasbeg)
	        x2 = 0;		// same as (x2     = x + rasbeg)
	    } else {
	        rasbeg = 0;
	        // width -= 0;	// same as (width -= rasbeg)
	        x2 = x;		// same as (x2     = x + rasbeg)
	    }
	    // rasline[rasbeg]          == pixel at coordinate (x2,y)
	    // rasline[width]           == pixel at coordinate (x+width, y)
	    // rasline[rasbeg + width]  == pixel at coordinate (x2+width, y)
	    if (x2 + width > global_width) {
	        width = global_width - x2;
	    }
	    if (width <= 0) {
	        return 1;
	    }
	    rasend = rasbeg + width;
	    // rasline[rasbeg] == pixel at coordinate (x2,y)
	    // rasline[rasend] == pixel at coordinate (x2+width, y)
	    int off = y * global_width + x2;
	    if (trans_pixel >= 0 && !firstframe) {
	        if (saved_image != null && saved_model == model) {
		        for (int i = rasbeg; i < rasend; i++, off++) {
		            byte pixel = rasline[i];
		            if ((pixel & 0xff) == trans_pixel) {
    			        rasline[i] = saved_image[off];
		            } else if (disposal_method == DISPOSAL_SAVE) {
	    		        saved_image[off] = pixel;
		            }
		        }
	        } else {
		    // We have to do this the hard way - only transmit
		    // the non-transparent sections of the line...
		    int runstart = -1;
		    int count = 1;
		    for (int i = rasbeg; i < rasend; i++, off++) {
		        byte pixel = rasline[i];
		        if ((pixel & 0xff) == trans_pixel) {
			    if (runstart >= 0) {
			        count = source.setPixels(x + runstart, y,
						         i - runstart, height,
						         model, rasline,
						         runstart, 0);
			        if (count == 0) {
				    break;
			        }
			    }
			    runstart = -1;
		        } else {
			    if (runstart < 0) {
			        runstart = i;
			    }
			    if (disposal_method == DISPOSAL_SAVE) {
			        saved_image[off] = pixel;
			    }
		        }
		    }
		    if (runstart >= 0) {
		        count = source.setPixels(x + runstart, y,
					         rasend - runstart, height,
					         model, rasline,
					         runstart, 0);
		    }
		    // Since (saved_model != model), store must be null...
		    return count;
	        }
	    } else if (disposal_method == DISPOSAL_SAVE) {
	        System.arraycopy(rasline, rasbeg, saved_image, off, width);
	    }
	    int count = source.setPixels(x2, y, width, height, model,
				         rasline, rasbeg, 0);
	    if (store != null
	        && store.setPixels(x2, y, width, height, rasline, rasbeg, 0))
	    {
	        count++;
	    }
	    return count;
    }

*/

    // Do source.setPixels, store.setPixels and copy into saved_image



    /**
     * Read Image data--called from produceImage()
     */
    private boolean readImage(boolean first) throws IOException {
	    long tm = 0;

	    if (verbose) {
	        tm = System.currentTimeMillis();
	    }

	    // Allocate the buffer
	    byte block[] = new byte[256 + 3];

/*
	    // Read the image descriptor
	    if (readBytes(block, 0, 10) != 0) {
	        throw new IOException();
	    }
	    int x = ExtractWord(block, 0);
	    int y = ExtractWord(block, 2);
	    int width = ExtractWord(block, 4);
	    int height = ExtractWord(block, 6);
	    boolean interlace = (block[8] & INTERLACEMASK) != 0;

	    IndexColorModel model = global_model;

	    if ((block[8] & COLORMAPMASK) != 0) {
	        // We read one extra byte above so now when we must
	        // transfer that byte as the first colormap byte
	        // and manually read the code size when we are done
	        int num_local_colors = 1 << ((block[8] & 0x7) + 1);

	        // Read local colors
	        byte[] local_colormap = new byte[num_local_colors * 3];
	        local_colormap[0] = block[9];
	        if (readBytes(local_colormap, 1, num_local_colors * 3 - 1) != 0) {
    		    throw new IOException();
	        }

	        // Now read the "real" code size byte which follows
	        // the local color table
	        if (readBytes(block, 9, 1) != 0) {
    		    throw new IOException();
	        }
	        model = new IndexColorModel(8, num_local_colors, local_colormap,
					    0, false, trans_pixel);
	    } else if (model == null
		       || trans_pixel != model.getTransparentPixel()) {
	        model = new IndexColorModel(8, num_global_colors, global_colormap,
					    0, false, trans_pixel);
	        global_model = model;
	    }
*/
	    // Notify the consumers
	    if (first) {
	        makeStore(global_width, global_height);
	        if (store != null) {
		        store.setProperties(props);
		        store.setColorModel(global_model);
	        }

	        source.setDimensions(global_width, global_height);
	        source.setProperties(props);
	        source.setColorModel(global_model);
	    } 
/*
        else if (store != null) {
	        if (model != store.getColorModel()) {
		        cancatchup = false;
		        store = null;
	        }
	    }
*/
	    if (disposal_method == DISPOSAL_SAVE && saved_image == null) {
	        saved_image = new byte[global_width * global_height];
	    }

//	    int hints = (interlace ? interlaceflags : normalflags);
	    source.setHints(hintsflags);
	    if (store != null) {
	        store.setHints(hintsflags);
	    }

	    // allocate the raster data
	    byte rasline[] = new byte[global_width];

	    if (verbose) {
	        System.out.print("Reading a " + global_width + " by " + global_height + " " 
		          + "image...");
	    }


        // Here's where I stick in my code to read
/*
	    boolean ret = parseImage(x, y, width, height,
				     interlace, ExtractByte(block, 9),
				     block, rasline, model);
*/

        int length = 0, val = 0, temp;
        byte[] buf = new byte[4];
        for (int row=0; row < global_height; row++) {  // for each row
            for (int col=0; col < global_width; col+= length) {    // for each column
                switch (tag) {
                  case TAG4:
                    readBytes(buf, 0, 1);
                    temp = ExtractByte(buf, 0);
                    length = temp >>> 4;
                    if (length == 0) length = 16;
                    val = temp & 0xf;
                    break;

                  case TAG8:
                    readBytes(buf, 0, 2);
                    temp = ExtractWord(buf, 0);
                    length = temp & 0xff;
                    if (length == 0) length = 256;
                    val = temp >>> 8;
                    break;

                  case TAG12:
                    readBytes(buf, 0, 2);
                    temp = ExtractWord(buf, 0);
                    length = temp >>> 12;
                    if (length == 0) length = 16;
                    val = temp & 0xfff;
                    break;

                  case TAG16:
                    readBytes(buf, 0, 3);
                    length = ExtractByte(buf, 0);
                    if (length == 0) length = 256;
                    val = ExtractWord(buf, 1);
                    break;

                  default:
                    System.err.println("Unknown tag type " + tag);
                } // end switch

                // Convert to color here
                byte storedVal = (byte)(val % 256);

                for (int j = 0; j < length; j++) 
                    rasline[col + j] = storedVal;
            }   // end for each column

            // Send one row of pixels at a time;
            source.setPixels(0, row, global_width, 1, global_model, rasline, 
                global_width, global_width);
            if (store != null) {
                store.setPixels(0, row, global_width, 1, rasline, 
                    global_width, global_width);
            }
            if (disposal_method == DISPOSAL_SAVE) {
	            System.arraycopy(rasline, 0, saved_image, row * global_width, 
                    global_width);
            }

        }   // end for each row


/*
	    if (ret) {
	        source.imageComplete(ImageConsumer.SINGLEFRAMEDONE);
	        if (delay > 0) {
		        try {
		            if (verbose) {
    			        System.out.println("sleeping: "+delay);
		            }
		            Thread.sleep(delay);
		        } catch (InterruptedException e) {
		            return false;
		        }
	        }
*/
	        if (verbose && disposal_method != 0) {
    		    System.out.println("disposal method: "+ disposal_method);
	        }
/*
	        if (x < 0) {
		        width += x;
		        x = 0;
	        }
	        if (x + width > global_width) {
    		    width = global_width - x;
	        }
	        if (width <= 0) {
    		    disposal_method = DISPOSAL_NONE;
	        } else {
		        if (y < 0) {
		            height += y;
		            y = 0;
		        }
		        if (y + height > global_height) {
		            height = global_height - y;
		        }
		        if (height <= 0) {
		            disposal_method = DISPOSAL_NONE;
		        }
	        }

	        switch (disposal_method) {

              case DISPOSAL_PREVIOUS:
		        if (saved_image != null) {
		            source.setPixels(x, y, width, height,
				             saved_model, saved_image,
				             y * global_width + x, global_width);
		            if (store != null) {
			        store.setPixels(x, y, width, height, saved_image,
					        y * global_width + x, global_width);
		            }
		        }
		        break;

	          case DISPOSAL_BGCOLOR:
		        byte tpix;
		        if (model.getTransparentPixel() < 0) {
		            cancatchup = false;
		            store = null;
		            model = trans_model;
		            if (model == null) {
			        model = new IndexColorModel(8, 1,
						            new byte[4], 0, true);
			        trans_model = model;
		            }
		            tpix = 0;
		        } else {
		            tpix = (byte) model.getTransparentPixel();
		        }
		        for (int i = 0; i < width; i++) {
		            rasline[i] = tpix;
		        }
		        source.setPixels(x, y, width, height, model, rasline, 0, 0);
		        if (store != null) {
		            store.setPixels(x, y, width, height, rasline, 0, 0);
		        }
		        break;

	          case DISPOSAL_SAVE:
		        saved_model = model;
		        break;
	        }   // end switch
	    }       // end if ret
*/
	    if (verbose) {
	        System.out.println("done in "
			           + (System.currentTimeMillis() - tm)
			           + "ms");
	    }


//	    return ret;
        return true;
    }
}
